from typing import Iterable

import pandas as pd

from ..metadata_reader import Defaults


def weighted_average(
    table: pd.DataFrame,
    defaults: Defaults,
    weight_col: str | None = None,
    columns: str | Iterable[str] | None = None,
) -> pd.Series:
    """Calculate the weighted average of columns in a DataFrame.

    Parameters
    ----------
    table : DataFrame
        Input DataFrame containing the columns to calculate
        weighted average for and the weight column.
    weight_col : str, default 'Weight'
        The name of the column containing the weights.

    Returns
    -------
    pandas Series
        A Series containing the weighted average of each column.

    Raises
    ------
    ValueError
        If the weight column is not in the table.

    Examples
    --------
    >>> df = pd.DataFrame({
            'Col1': [1, 2, 3],
            'Col2': [25, 15, 10],
            'Weight': [0.25, 0.25, 0.5]})
    >>> weighted_average(df)
        Col1     2.25
        Col2    15.00
        dtype: float64

    """
    weight_col = defaults.columns.weight if weight_col is None else weight_col
    if weight_col not in table.columns:
        raise ValueError(f"Weight column {weight_col} not in table")
    if columns is None:
        columns = [
            col
            for col in table.select_dtypes("number").columns
            if col in table.columns
            if col not in defaults.columns.groupby
            if col not in [defaults.columns.id, weight_col, "index"]
        ]
    elif isinstance(columns, str):
        columns = [columns]
    else:
        columns = list(columns)
    calc_table = table.loc[:, columns + [weight_col]].copy()
    calc_table[columns] = calc_table[columns].multiply(
        calc_table[weight_col], axis="index"
    )
    columns_summation = calc_table.sum()
    results = (
        columns_summation.loc[columns]
        .divide(columns_summation[weight_col])
        .loc[columns]
    )
    return results


def average_table(
    table: pd.DataFrame,
    defaults: Defaults,
    columns: list[str] | None = None,
    groupby: list[str] | str | None = None,
    weight_col: str | None = None,
) -> pd.DataFrame:
    if isinstance(table.columns, pd.MultiIndex):
        is_multi_index = True
        column_names = table.columns.names
        table.columns = table.columns.to_flat_index()
    else:
        is_multi_index = False
        column_names = None

    table = table.reset_index().copy()

    if groupby is None:
        groupby = [col for col in table.columns if col in defaults.columns.groupby]
    elif isinstance(groupby, str):
        groupby = [groupby]
    if groupby is None:
        groupby = []

    weight_col = defaults.columns.weight if weight_col is None else weight_col

    if len(groupby) == 0:
        row = weighted_average(
            table, columns=columns, weight_col=weight_col, defaults=defaults
        )
        result = pd.DataFrame([row])
    else:
        result = table.groupby(groupby).apply(
            weighted_average, columns=columns, weight_col=weight_col, defaults=defaults
        )

    if is_multi_index:
        result.columns = pd.MultiIndex.from_tuples(result.columns, names=column_names)

    return result



import pandas as pd


def collapse_years(table: pd.DataFrame) -> pd.DataFrame:
    table = table.copy()
    last_year = table.index.to_list()[-1]
    filt = -(table == table.shift(1)).all(axis="columns")
    table = table.loc[filt]
    start_year = table.index.to_series()
    end_year = start_year.shift(-1, fill_value=last_year + 1) - 1
    new_index = start_year.astype(str)
    filt = start_year != end_year
    new_index.loc[filt] = new_index.astype(str) + "-" + end_year.astype(str)
    table.index = pd.Index(new_index, name="Years")
    return table


def maybe_to_numeric(column: pd.Series) -> pd.Series:
    try:
        column = column.str.replace(r"\s+", "", regex=True).astype("Float64")
    except ValueError:
        pass
    try:
        column = column.astype("Int64")
    except ValueError:
        pass
    return column



import pandas as pd

from ..api import API

from .quantile import QuantileSettings, Quantiler

from .average import weighted_average, average_table

# pylint: disable=too-many-arguments
# pylint: disable=unused-argument


class Calculator:
    def __init__(self, api: API) -> None:
        self._api = api

    def weighted_average(
        self,
        table: pd.DataFrame,
        weight_col: str | None = None,
        columns: list[str] | None = None,
    ) -> pd.Series:
        return weighted_average(
            table=table,
            defaults=self._api.defaults,
            columns=columns,
            weight_col=weight_col,
        )

    def average_table(
        self,
        table: pd.DataFrame,
        weight_col: str | None = None,
        groupby: list[str] | str | None = None,
        columns: list[str] | None = None,
    ) -> pd.DataFrame:
        return average_table(
            table=table,
            defaults=self._api.defaults,
            columns=columns,
            groupby=groupby,
            weight_col=weight_col,
        )

    def quantile(
        self,
        *,
        table: pd.DataFrame | pd.Series | None = None,
        quantile_column_name: str = "Quantile",
        bins: int = -1,
        **kwargs
    ) -> pd.Series:
        settings = QuantileSettings(api=self._api, **kwargs)
        quantile = Quantiler(table=table, settings=settings).calculate_quantile()
        quantile = quantile.rename(quantile_column_name)
        if bins > 0:
            quantile = (
                quantile.multiply(bins)
                .floordiv(1)
                .add(1)
                .clip(1, bins)
                .astype(int)
                .rename(quantile_column_name)
            )
        return quantile

    def add_quantile(
        self, table: pd.DataFrame, quantile_column_name: str = "Quantile", **kwargs
    ) -> pd.DataFrame:
        kwargs.update({"quantile_column_name": quantile_column_name})
        table[quantile_column_name] = self.quantile(table=table, **kwargs)
        return table

    def add_decile(
        self, table: pd.DataFrame, quantile_column_name: str = "Decile", **kwargs
    ) -> pd.DataFrame:
        kwargs.update({"bins": 10, "quantile_column_name": quantile_column_name})
        table[quantile_column_name] = self.quantile(table=table, **kwargs)
        return table

    def add_percentile(
        self, table: pd.DataFrame, quantile_column_name: str = "Percentile", **kwargs
    ) -> pd.DataFrame:
        kwargs.update({"bins": 100, "quantile_column_name": quantile_column_name})
        table[quantile_column_name] = self.quantile(table=table, **kwargs)
        return table



"""Decodes household data using metadata mappings.

Provides functionality to resolve metadata versions, map codes 
to metadata attributes, and join decoded attributes onto data tables.

Classes
-------
CommodityDecoderSettings - Settings for commodity code decoding.
CommodityDecoder - Decodes commodity codes using metadata.
IDDecoderSettings - Settings for household ID decoding.  
IDDecoder - Decodes household IDs using metadata.

Functions
--------- 
read_classification_info - Reads classified metadata by name.
create_classification_table - Creates table from classified metadata.

The decoders resolve metadata versions, map codes to attributes, 
and add decoded columns to the input tables.

The functions provide helpers for resolving metadata and reading
classification info from the raw metadata.

"""
from itertools import product
from typing import Callable, Iterable, Literal, Annotated, Any, Optional

import pandas as pd
from pydantic import BaseModel, ConfigDict, BeforeValidator, Field

from .metadata_reader import Defaults, Metadata

from . import utils, external_data


def maybe_to_tuple(_input: Any) -> tuple:
    """Converts input to a tuple if needed.

    Parameters
    ----------
    _input : Any
        Input to convert.

    Returns
    -------
    tuple
        Input converted to a tuple.

    Notes
    -----
    If _input is already a tuple, returns it unchanged.
    If _input is a single value, returns it in a 1-tuple.
    If _input is an iterable, converts it to a tuple.
    """
    if isinstance(_input, tuple):
        return _input
    if isinstance(_input, str):
        return (_input,)
    if isinstance(_input, Iterable):
        return tuple(_input)
    return (_input,)


def extract_column(table: pd.DataFrame, column_name: str) -> pd.Series:
    """Extracts a column from a DataFrame as a Series.

    Checks table columns, index name(s) for the given column
    name and returns the matching column as a Series.

    Raises KeyError if column not found.

    Parameters
    ----------
    table : DataFrame
        DataFrame to extract column from.

    column_name : str
        Name of column to extract.

    Returns
    -------
    Series
        Extracted column as a Series.

    Raises
    ------
    KeyError
        If column not found in DataFrame.

    """
    if column_name in table.columns:
        column = table.loc[:, column_name].copy()
    elif isinstance(table.index, pd.Index) and table.index.name == column_name:
        column = table.index.to_series()
    elif isinstance(table.index, pd.MultiIndex) and column_name in table.index.names:
        column = table.index.to_frame().loc[:, column_name].copy()
    else:
        raise KeyError
    return column


_Aspects = Annotated[tuple[str, ...], BeforeValidator(maybe_to_tuple)]
_Levels = Annotated[tuple[int, ...], BeforeValidator(maybe_to_tuple)]
_ColumnNames = Annotated[tuple[str, ...], BeforeValidator(maybe_to_tuple)]


class DecoderSettings(BaseModel):
    """Settings for decoding classification codes.

    Attributes
    ----------
    target : str
        Column to decode.

    classification_type : Literal["commodity", "occupation", "industry"], optional
        Type of classification. Inferred automatically if not provided.

    name : str, default "original"
        Name of classification metadata to use.

    year_col : str, optional
        Column name for year. Inferred automatically if not provided.

    versioned_info : dict
        Classification metadata. Populated automatically based on name and
        classification_type.

    defaults : dict
        Default settings. Merged with configured settings.

    aspects : tuple[str]
        Labels to extract as output columns.

    levels : tuple[int]
         Hierarchy levels to extract.

    drop_value : bool, default False
        Whether to drop rows when value is missing.

    column_names : tuple[str]
        Names for extracted output columns.

    missing_value_replacements : dict
        Replacements for missing values in output columns.

    See Also
    --------
    Decoder : Uses these settings to decode codes.

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    lib_defaults: Defaults
    lib_metadata: Metadata
    target: str
    classification_type: Optional[Literal["commodity", "occupation", "industry"]] = None
    name: str = "original"
    year_col: str = Field(None, validate_default=False)
    versioned_info: dict = {}
    defaults: dict = {}
    aspects: _Aspects = ()
    levels: _Levels = ()
    drop_value: bool = False
    column_names: _ColumnNames = ()
    missing_value_replacements: dict[str, str] = {}

    def model_post_init(self, __contex=None) -> None:
        if self.classification_type is None:
            self.classification_type = self._resolve_classification_type()

        if self.year_col is None:
            self.year_col = self.lib_defaults.columns.year

        if self.classification_type == "commodity":
            self.versioned_info = self.lib_metadata.commodities[self.name]
        elif self.classification_type == "industry":
            self.versioned_info = self.lib_metadata.industries[self.name]
        elif self.classification_type == "occupation":
            self.versioned_info = self.lib_metadata.occupations[self.name]
        else:
            raise ValueError(f"{self.classification_type} is not valid type")

        if "defaults" in self.versioned_info:
            self.defaults = self.versioned_info["defaults"]
        for key, value in self.defaults.items():
            if isinstance(value, list):
                value = tuple(value)
            if (getattr(self, key) is None) or (len(getattr(self, key)) == 0):
                setattr(self, key, value)
        if len(self.aspects) == 0:
            self.aspects = ("item_key",)
        if len(self.levels) == 0:
            self.levels = (1,)
        self._resolve_column_names()

    def _resolve_classification_type(
        self,
    ) -> Literal["commodity", "occupation", "industry"]:
        """Resolve the classification type based on keywords."""
        for name, keywords in [
            ("commodity", self.lib_defaults.columns.commodity_code),
            ("industry", self.lib_defaults.columns.industry_code),
            ("occupation", self.lib_defaults.columns.occupation_code),
        ]:
            for keyword in keywords:
                if keyword.lower() in self.target.lower():
                    return name
        raise ValueError("Cannot specify classification type")

    def _resolve_column_names(self) -> None:
        if len(self.column_names) == 0:
            names = [
                f"{label}_{level}"
                for label, level in product(self.aspects, self.levels)
            ]
            self.column_names = tuple(names)
        elif len(self.column_names) == len(self.aspects) * len(self.levels):
            pass
        elif len(self.column_names) == len(self.aspects):
            names = [
                f"{label}_{level}"
                for label, level in product(self.column_names, self.levels)
            ]
            self.column_names = tuple(names)

    @property
    def rename_dict(self):
        """Mapping of label-level keys to output column names.

        Returns a dictionary mapping each combination of labels
        and levels to the corresponding output column name.

        This can be used to rename the multi-level columns in the
        decoded mapping table to the configured output names.

        Returns
        -------
        dict
            Mapping of label-level tuples to output column names.

        """
        label_level = product(self.aspects, self.levels)
        return dict(zip(label_level, self.column_names))


class Decoder:
    """Decodes commodity codes using classification metadata.

    Parameters
    ----------
    table : DataFrame
        Table with code and year columns to decode.

    settings : DecoderSettings
        Decoding configuration settings.

    Attributes
    ----------
    classification_table : DataFrame
        Resolved classification metadata.

    year_code_pairs : DataFrame
        Unique year and code combinations.

    Methods
    -------
    create_mapping_table()
        Maps codes to metadata based on year.

    add_classification()
        Adds decoded columns to the input table.

    See Also
    --------
    DecoderSettings : Decoding configuration.

    """

    def __init__(self, table: pd.DataFrame, settings: DecoderSettings) -> None:
        self.table = table
        self.settings = settings
        self.code_column = extract_column(table, settings.target)
        self.year_column = extract_column(table, settings.year_col)
        self.classification_table = self.create_classification_table(
            years=self.year_column.drop_duplicates().to_list(),
        )
        self.year_code_pairs = self._create_year_code_pairs()

    def create_classification_table(
        self,
        years: Iterable[int],
    ) -> pd.DataFrame:
        """Creates classification table for given years.

        Loops through the provided years, reads the classification metadata,
        converts to annual DataFrames, concatenates the results, and returns
        the final classification table.

        Parameters
        ----------

        years : Iterable[int]
            Years to include in the resulting table.

        Returns
        -------
        DataFrame
            Classification table with a row for each year.

        See Also
        --------
        _create_annual_classification_table : Converts metadata to DataFrame.

        """
        table_list = []
        for year in years:
            classification_info = utils.resolve_metadata(
                self.settings.versioned_info, year, categorize=True
            )
            assert isinstance(classification_info, dict)
            annual_table = pd.DataFrame(classification_info.get("items"))
            annual_table["code_range"] = annual_table["code"].apply(
                utils.Argham,  # type: ignore
                default_start=self.settings.lib_defaults.years[0],
                default_end=self.settings.lib_defaults.years[-1] + 1,
                keywords=["code"],
            )
            annual_table = annual_table.drop(columns=["code"])
            annual_table.loc[:, "Year"] = year
            table_list.append(annual_table)
        table = pd.concat(table_list, ignore_index=True)
        return table

    def _create_year_code_pairs(self) -> pd.DataFrame:
        years = self.year_column.drop_duplicates()
        yc_pair_list = []
        for year in years:
            filt = self.year_column == year
            codes = self.code_column.loc[filt].drop_duplicates()
            yc_pair = codes.to_frame()
            yc_pair[self.settings.year_col] = year
            yc_pair_list.append(yc_pair)
        return pd.concat(yc_pair_list, ignore_index=True)

    def _build_year_code_table(
        self, year_code_pairs: pd.DataFrame, row: pd.Series
    ) -> pd.DataFrame:
        filt = (
            year_code_pairs[self.settings.target]
            .apply(lambda x: x in row["code_range"])
        )
        filt = filt & (
            year_code_pairs[self.settings.year_col] == row[self.settings.year_col]
        )
        matched_codes = year_code_pairs.loc[filt].set_index(
            [self.settings.year_col, self.settings.target]
        )
        columns = row.drop(["code_range", self.settings.year_col]).index
        code_table = pd.DataFrame(
            data=[row.loc[columns]] * len(matched_codes.index),
            index=matched_codes.index,
            columns=columns,
        )
        return code_table

    def create_mapping_table(self) -> pd.DataFrame:
        """Creates code mapping table from metadata.

        Loops through classification table and builds a mapping
        table linking codes to metadata based on year.

        Multi-index columns are renamed using the settings.
        Table is validated before returning.

        Returns
        -------
        DataFrame
            Mapping table with year, code and decoded columns.

        See Also
        --------
        _build_year_code_table : Builds table for a single row.

        _validate_mapping_table : Validates mapping integrity.

        """
        code_table_list = []
        for _, row in self.classification_table.iterrows():
            code_table = self._build_year_code_table(self.year_code_pairs, row)
            if not code_table.empty:
                code_table_list.append(code_table)
        mapping_table = pd.concat(code_table_list)
        mapping_table = mapping_table.set_index("level", append=True)
        self._validate_mapping_table(mapping_table)
        mapping_table = mapping_table.unstack(-1)
        mapping_table = mapping_table.loc[:, self.settings.rename_dict.keys()]  # type: ignore
        mapping_table.columns = self.settings.rename_dict.values()
        return mapping_table

    def _validate_mapping_table(self, mapping_table: pd.DataFrame):
        filt = mapping_table.index.duplicated(keep=False)
        if filt.sum() > 0:
            invalid_case_sample = (
                mapping_table.loc[filt]
                .sort_values([self.settings.target, "level"])
                .head(10)
            )
            raise ValueError(f"Classification is not valid \n{invalid_case_sample}")

    def _fill_missing_values(self):
        if "missing_value_replacements" not in self.settings.defaults:
            return
        for column, default in self.settings.defaults[
            "missing_value_replacements"
        ].items():
            if column not in self.table.columns:
                continue
            filt = self.table.loc[:, column].isna()
            self.table.loc[filt, column] = default  # type: ignore

    def add_classification(self):
        """Adds decoded columns to the input table.

        Joins the mapping table to the input table using
        the year and code columns.

        Also fills in any missing values based on defaults.

        Returns
        -------
        DataFrame
            Input table with decoded columns added.

        """
        mapping = self.create_mapping_table()
        self.table = self.table.join(
            mapping, on=[self.settings.year_col, self.settings.target]
        )
        self._fill_missing_values()
        return self.table


class IDDecoderSettings(BaseModel):
    """Settings for decoding household IDs.

    Attributes
    ----------
    name : Attribute
        Name of household attribute to decode.

    id_col : str
        Column name for household IDs.

    year_col : str
        Column name for year.

    fields : tuple[str]
        Labels to extract as output columns.

    column_names : tuple[str]
        Names of columns to add to output table.

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    aspects: _Aspects = ("name",)
    column_names: _ColumnNames = ()
    lib_defaults: Defaults
    lib_metadata: Metadata

    id_col: str | None = None
    year_col: str | None = None

    def model_post_init(self, __contex=None) -> None:
        if self.id_col is None:
            self.id_col = self.lib_defaults.columns.id
        if self.year_col is None:
            self.year_col = self.lib_defaults.columns.year
        self._resolve_column_names()

    def _resolve_column_names(self) -> None:
        if len(self.column_names) != len(self.aspects):
            if len(self.aspects) == 1:
                names = [self.name]
            else:
                names = [f"{self.name}_{label}" for label in self.aspects]
            self.column_names = tuple(names)


class IDDecoder:
    """Decodes household IDs using metadata mappings.

    Parameters
    ----------
    table : DataFrame
        Table with ID and year columns to decode.

    settings : IDDecoderSettings
        Configuration settings for decoding.

    Methods
    -------
    construct_mapping_table()
        Builds mapping table from metadata.

    add_attribute()
        Adds decoded columns to the input table.
    """

    def __init__(
        self,
        table: pd.DataFrame,
        settings: IDDecoderSettings,
    ) -> None:
        self.table = table
        self.settings = settings
        self.id_series = extract_column(table, settings.id_col)
        self.year_series = extract_column(table, settings.year_col)

    def construct_mapping_table(self) -> pd.DataFrame:
        """Constructs metadata mapping table for household IDs.

        Maps ID column to decoded labels based on year. Concatenates
        the ID, year, and label columns into a mapping table.

        Returns
        -------
        DataFrame
            Mapping table with year, ID and decoded columns.

        """
        mapped_columns = [self.year_series, self.id_series]
        for label in self.settings.aspects:
            mapped_column = self._map_id_to_label(label)
            mapped_columns.append(mapped_column)
        year_and_id = [self.settings.year_col, self.settings.id_col]
        columns = year_and_id + list(self.settings.column_names)
        mapping_table = pd.concat(mapped_columns, axis="columns", keys=columns)
        mapping_table = mapping_table.drop_duplicates().set_index(year_and_id)
        return mapping_table

    def _create_code_builder(
        self, household_metadata: dict
    ) -> Callable[[pd.Series], pd.Series]:
        ld_len = household_metadata["ID_Length"]
        attr_dict = household_metadata[self.settings.name]["code"]

        if ("position" in attr_dict) and attr_dict["position"] is not None:
            start, end = attr_dict["position"]["start"], attr_dict["position"]["end"]

            def builder(household_id_column: pd.Series) -> pd.Series:
                return (
                    household_id_column
                    % pow(10, (ld_len - start))
                    // pow(10, (ld_len - end))
                )

        elif "external_file" in attr_dict:
            file_name = attr_dict["external_file"]
            code_builer_file = external_data.load_table(
                file_name, self.settings.lib_defaults, reset_index=False
            )
            code_series = code_builer_file.loc[household_metadata["year"]].iloc[:, 0]
            assert isinstance(code_series, pd.Series)
            mapping_dict = code_series.to_dict()

            def builder(household_id_column: pd.Series) -> pd.Series:
                codes = household_id_column.map(mapping_dict)
                assert codes.isna().sum() == 0
                return codes

        else:
            raise ValueError("Code position is not available")

        return builder

    def _create_code_mapper(
        self, label: str, year: int
    ) -> Callable[[pd.Series], pd.Series]:
        id_metadata = utils.resolve_metadata(
            self.settings.lib_metadata.id_information, year
        )

        if label == "code":
            return self._create_code_builder(id_metadata)

        if not isinstance(id_metadata, dict):
            raise ValueError
        # pylint: disable=unsubscriptable-object
        if label in id_metadata[self.settings.name].get("mappings", {}):
            post_mapping_info = id_metadata[self.settings.name]["mappings"][label]
            label = post_mapping_info["origin"]
            post_mapping = self.settings.lib_metadata.id_information[
                post_mapping_info["mapping"]
            ]
        else:
            post_mapping = None
        mapping = id_metadata[self.settings.name][label]
        code_builder = self._create_code_builder(id_metadata)

        def mapper(household_id_column: pd.Series) -> pd.Series:
            mapped = code_builder(household_id_column).map(mapping)
            if post_mapping is not None:
                mapped = mapped.map(post_mapping)
            mapped = mapped.astype("category")
            mapped.name = label
            return mapped

        return mapper

    def _map_id_to_label(self, label: str):
        years = self.year_series.drop_duplicates()
        attribute_column = pd.Series(index=self.table.index, dtype="object")
        for year in years:
            filt = self.year_series == year
            attribute_column.loc[filt] = self._create_code_mapper(label, year)(
                self.id_series.loc[filt]
            )
        return attribute_column

    def add_attribute(self):
        """Adds decoded columns to the input table.

        Joins the mapping table to the input table using the ID and
        year columns.

        Returns
        -------
        DataFrame
            Input table with decoded columns added.

        """
        mapping_table = self.construct_mapping_table()
        year_and_id = [self.settings.year_col, self.settings.id_col]
        self.table = self.table.join(mapping_table, year_and_id)
        return self.table



import pandas as pd


def create_year_month_index(from_year: int, to_year: int) -> pd.MultiIndex:
    return pd.MultiIndex.from_product(
        [range(from_year, to_year + 1), range(1, 12 + 1)], names=["Year", "Month"]
    )


def sci_cpi_1395_urban_singleindex_monthly(table: pd.DataFrame) -> pd.DataFrame:
    table = table.loc[2:, [2]]
    table.index = create_year_month_index(1361, 1401)
    table.columns = ["CPI"]
    return table


def sci_cpi_1395_urban_singleindex_annual(table: pd.DataFrame) -> pd.DataFrame:
    table.columns = ["Year", "CPI"]
    table = table.loc[2:]
    table = table.set_index("Year")
    return table


def sci_cpi_1395_rural_maingroups_monthly(table: pd.DataFrame) -> pd.DataFrame:
    table = table.loc[[3], 53:].T
    table.index = create_year_month_index(1374, 1401)
    table.columns = ["CPI"]
    return table


def sci_cpi_1395_rural_maingroups_annual(table: pd.DataFrame) -> pd.DataFrame:
    table = table.loc[[4], 1:].T
    table.columns = ["CPI"]
    table.index = pd.Index(range(1361, 1401), name="Year")
    return table


def sci_cpi_1395_monthly(table: pd.DataFrame) -> pd.DataFrame:
    table = table.loc[[3], 1:].T
    table.columns = ["CPI"]
    table.index = create_year_month_index(1390, 1401)
    return table


def sci_cpi_1395_annual(table: pd.DataFrame) -> pd.DataFrame:
    table = table.loc[[4], 1:].T.astype("float64")
    table.columns = ["CPI"]
    table.index = pd.Index(range(1390, 1401), name="Year")
    return table


def sci_cpi_1395_monthly_urban_rural(table_list: list[pd.DataFrame]) -> pd.DataFrame:
    return pd.concat(
        table_list, keys=["Urban", "Rural"], names=["Urban_Rural", "Year", "Month"]
    )


def sci_cpi_1395_annual_urban_rural(table_list: list[pd.DataFrame]) -> pd.DataFrame:
    return pd.concat(table_list, keys=["Urban", "Rural"], names=["Urban_Rural", "Year"])


def sci_cpi_1400_urban_singleindex_monthly(table: pd.DataFrame) -> pd.DataFrame:
    table = table.loc[2:, [2]]
    table.index = create_year_month_index(1361, 1402)
    table.columns = ["CPI"]
    return table


def sci_cpi_1400_urban_singleindex_annual(
    table_list: list[pd.DataFrame],
) -> pd.DataFrame:
    table = table_list[0].groupby("Year")[["CPI"]].mean()
    return table


def sci_cpi_1400_rural_maingroups_monthly(table: pd.DataFrame) -> pd.DataFrame:
    table.loc[1, :] = table.loc[1].ffill()
    table.loc[1, 0] = "Year"
    table.loc[2, 0] = "Month_Seasion"
    table = table.loc[1:]
    table = table.set_index(0)
    table.index.name = None
    table = table.T
    table["Year"] = table["Year"].astype(int)
    table = table.set_index(["Year", "Month_Seasion"])
    table = table.replace(r"[\s\-]", None, regex=True)
    return table


def sci_cpi_1400_rural_maingroups_annual(
    table_list: list[pd.DataFrame],
) -> pd.DataFrame:
    table = table_list[0].groupby("Year").mean()
    return table


def sci_cpi_1400_annual_urban_rural(
    table_list: list[pd.DataFrame],
) -> pd.DataFrame:
    urban_table, rural_table = table_list
    rural_table = rural_table.iloc[:, [0]]
    rural_table.columns = ["CPI"]

    table = pd.concat(
        [urban_table, rural_table],
        keys=["Urban", "Rural"],
        names=["Urban_Rural", "Year"],
    )
    return table


def sci_gini_annual(table: pd.DataFrame) -> pd.DataFrame:
    table = table.loc[[2], 1:].T
    index = pd.Index(range(1363, 1402), name="Year")
    table = table.set_axis(index, axis="index").set_axis(["Gini"], axis="columns")
    return table



import subprocess
from pathlib import Path
import platform

from .download_utils import download_7zip
from ..metadata_reader import BASE_PACKAGE_DIRECTORY


def extract(
    compressed_file: Path,
    output_directory: Path,
    *,
    seven_zip_directory: Path = BASE_PACKAGE_DIRECTORY,
) -> None:
    """
    Extracts the contents of a compressed file using the 7-Zip tool.

    :param compressed_file_path: The path to the compressed file to be extracted.
    :type compressed_file_path: str
    :param output_directory: The path to the directory where the extracted files will be stored.
    :type output_directory: str
    :return: None, as this function does not return anything.

    """
    if platform.system() == "Windows":
        seven_zip_file_path = seven_zip_directory.joinpath("7-Zip", "7z.exe")
        if not seven_zip_file_path.exists():
            download_7zip()
        subprocess.run(
            [
                seven_zip_file_path,
                "e",
                compressed_file,
                f"-o{output_directory}",
                "-y",
            ],
            check=False,
            shell=True,
        )
    elif platform.system() == "Linux":
        seven_zip_file_path = seven_zip_directory.joinpath("7-Zip", "7zz")
        if not seven_zip_file_path.exists():
            download_7zip()
        subprocess.run(
            [
                seven_zip_file_path,
                "e",
                compressed_file,
                f"-o{output_directory}",
                "-y",
            ],
            check=False,
        )
    else:
        print("Your OS is not supported. try buy a new one.")



from pathlib import Path
import platform
from zipfile import ZipFile

import requests

from ..metadata_reader import defaults


def download(url: str, path: Path) -> None:
    """Downloads a file from a given URL and saves it to a specified local path.

    This function uses the requests library to send a GET request to the provided URL,
    and then writes the response content to a file at the specified path. If the path
    is not provided, the file is saved in a temporary directory. The function also
    provides an option to display a progress bar during the download.

    Parameters
    ----------
    url : str
        The URL of the file to download.
    path : str, Path, optional
        The local path where the downloaded file should be saved. If None, the file
        is saved in a temporary directory. Default is None.
    show_progress_bar : bool, optional
        If True, a progress bar is displayed during the download. Default is False.

    Returns
    -------
    Path
        The local path where the downloaded file was saved.

    Raises
    ------
    FileNotFoundError
        If the file cannot be found at the given URL.
    """
    response = requests.get(url, timeout=1000, stream=True)
    content_iterator = response.iter_content(chunk_size=4096)
    remote_file_size = response.headers.get("content-length")
    if remote_file_size is not None:
        remote_file_size = int(remote_file_size)
    else:
        raise FileNotFoundError("File is not found on the server")
    if path.exists():
        local_file_size = path.stat().st_size
    else:
        path.parent.mkdir(parents=True, exist_ok=True)
        local_file_size = 0
    if remote_file_size == local_file_size:
        return
    with open(path, mode="wb") as file:
        while True:
            try:
                chunk = next(content_iterator)
            except StopIteration:
                break
            except requests.Timeout:
                continue
            file.write(chunk)


def download_7zip():
    """
    Download the appropriate version of 7-Zip for the current operating system
    and architecture, and extract it to the root directory.

    """
    print(
        f"Downloading 7-Zip for {platform.system()} with "
        f"{platform.architecture()[0]} architecture"
    )
    file_name = f"{platform.system()}-{platform.architecture()[0]}.zip"
    file_path = defaults.base_package_dir.joinpath(file_name)

    if defaults.colab_mode:
        url = defaults.seven_zip_url
    else:
        url = f"{defaults.bucket_address}/7-Zip/{file_name}"
    download(url, file_path)

    with ZipFile(file_path) as zip_file:
        zip_file.extractall(defaults.base_package_dir)
    file_path.unlink()

    if platform.system() == "Linux":
        defaults.base_package_dir.joinpath("7-Zip", "7zz").chmod(0o771)


def download_map(
    map_name: str, source: str, *, map_metadata: dict, maps_directory: Path
) -> None:
    url = map_metadata[map_name][f"{source}_link"]
    file_path = maps_directory.joinpath("map.zip")
    download(url, file_path)
    path = maps_directory.joinpath(map_name)
    path.mkdir(exist_ok=True, parents=True)
    with ZipFile(file_path) as zip_file:
        zip_file.extractall(path)
    file_path.unlink()



from typing import Literal

import pandas as pd
from pydantic import BaseModel, ConfigDict


from ..api import API


_QuantileBase = Literal[
    "Income",
    "Expenditure",
    "Gross_Expenditure",
    "Net_Expenditure",
]


_EquivalenceScale = Literal[
    "Household", "Per_Capita", "OECD", "OECD_Modified", "Square_Root"
]


class QuantileSettings(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    api: API
    weight_column: str | None = None
    on_column: str | None = None
    on_variable: _QuantileBase | None = None
    default_variable: _QuantileBase = "Net_Expenditure"
    weighted: bool = True
    equivalence_scale: _EquivalenceScale = "Household"
    for_all: bool = True
    annual: bool = True
    groupby: list[str] = []
    years: list[int] | None = None

    def model_post_init(self, __context=None) -> None:
        if self.weight_column is None:
            self.weight_column = self.api.defaults.columns.weight


class Quantiler:
    """
    Quantiler Class
    """

    variable_aliases = {
        "Expenditure": "Gross_Expenditure",
    }

    variable_tables: dict[str, str] = {
        "Income": "Total_Income",
        "Gross_Expenditure": "Total_Expenditure",
        "Net_Expenditure": "Total_Expenditure",
    }

    def __init__(
        self,
        settings: QuantileSettings,
        table: pd.DataFrame | pd.Series | None = None,
    ) -> None:
        self.settings = settings.model_copy()

        self.table = table
        if self.table is not None:
            if ("Year" in self.table.index.names) and ("ID" in self.table.index.names):
                self.original_index = pd.DataFrame(index=self.table.index)
            else:
                self.original_index = self.table[["Year", "ID"]]
            self.table = self.table.copy().reset_index().set_index(["Year", "ID"])
        self.years = self._find_years()
        self.value_table = self.create_value_table()

    def create_value_table(self) -> pd.DataFrame:
        if self.settings.on_column is not None:
            if self.table is None:
                raise ValueError(
                    "The table must be provided for the `on_column` method"
                )
            value_table = self._extract_value_table(self.table)
        elif self.settings.on_variable is not None:
            value_table = self._get_external_value_table(self.settings.on_variable)
        else:
            value_table = self._get_external_value_table(self.settings.default_variable)

        equivalence_scale = (
            self.settings.api.load_table("Equivalence_Scale", years=self.years)
            .set_index(["Year", "ID"])
            .loc[:, self.settings.equivalence_scale]
        )
        value_table["Values"] = value_table["Values"].div(equivalence_scale)
        value_table = value_table.reset_index().dropna().sort_values("Values")
        return value_table

    def _find_years(self) -> list[int]:
        if self.settings.years is not None:
            return self.settings.years
        if self.table is None:
            raise ValueError("Year is Not available")
        if "Year" in self.table.index.names:
            return self.table.index.get_level_values("Year").unique().to_list()
        if "Year" in self.table.columns:
            return list(self.table["Year"].unique())
        raise ValueError("Year must be provided")

    def _extract_value_table(self, table: pd.DataFrame | pd.Series) -> pd.DataFrame:
        if isinstance(table, pd.Series):
            table = table.to_frame(name="Values")
            index_levels = list(table.index.names)
            index_levels.remove("Year")
            index_levels.remove("ID")
            table.index = table.index.droplevel(index_levels)  # type: ignore
        if isinstance(table, pd.DataFrame):
            table = table.reset_index().set_index(["Year", "ID"])
            if len(table.columns) > 1:
                if self.settings.on_column is None:
                    raise ValueError("Column name must be provided")
                assert self.settings.on_column in table.columns
                table = table[[self.settings.on_column]]
            table.columns = ["Values"]
        assert table.index.duplicated().sum() == 0
        return table

    def _get_external_value_table(self, variable: str) -> pd.DataFrame:
        variable = self.variable_aliases.get(variable, variable)
        table_name = self.variable_tables[variable]
        value_table = (
            self.settings.api.load_table(table_name, self.years)
            .set_index(["Year", "ID"])[[variable]]
            .rename(columns={variable: "Values"})
        )
        if (not self.settings.for_all) and (self.table is not None):
            value_table = value_table.loc[self.table.set_index(["Year", "ID"]).index]
        return value_table

    def calculate_quantile(self) -> pd.Series:
        groupby_columns = ["Year"] if self.settings.annual else []
        groupby_columns.extend(self.settings.groupby)
        quantile = (
            self.value_table.pipe(self._add_attributes)
            .pipe(self._add_weights)
            .groupby(groupby_columns)
            .apply(self._calculate_subgroup_quantile)
            .set_index(["Year", "ID"])["Quantile"]
        )
        quantile = self._align_with_table(quantile)
        return quantile

    def _calculate_subgroup_quantile(self, subgroup: pd.DataFrame) -> pd.DataFrame:
        return subgroup.assign(
            CumWeight=lambda df: df[self.settings.weight_column].cumsum(),
            Quantile=lambda df: df["CumWeight"] / df["CumWeight"].iloc[-1],
        ).loc[:, ["Year", "ID", "Quantile"]]

    def _add_weights(self, table: pd.DataFrame) -> pd.DataFrame:
        if self.settings.weighted:
            return self.settings.api.add_weight(table)
        return table.assign(Weight=1)

    def _add_attributes(self, table: pd.DataFrame) -> pd.DataFrame:
        for attribute in self.settings.groupby:  # type: ignore
            table = self.settings.api.add_attribute(table, name=attribute)
        return table

    def _align_with_table(self, quantile: pd.Series) -> pd.Series:
        if self.table is None:
            return quantile
        quantile = self.original_index.join(quantile, on=["Year", "ID"])["Quantile"]
        return quantile



from typing import Iterable

from pydantic import BaseModel


class ArghamDefaults(BaseModel):
    keywords: list[str] = []
    default_start: int | None = None
    default_end: int | None = None
    default_step: int = 1
    default_range: tuple[int, int] | None = None


class Argham:
    """Flexible argument handler for numbers and ranges.

    Accepts numbers, ranges, lists, and dicts to define a set
    of numbers and numeric ranges. Parses the input into
    number_list and range_list.

    Implements membership testing via 'in' operator based
    on the contained numbers and ranges.

    Can be used to flexibly defined sets of numbers/ranges
    and perform set-like membership tests on them.

    Parameters
    ----------
    argham : list, dict, int
        Input content. Can mix numbers, ranges, lists, and dicts.

    keywords : list of str, optional
        Keys in dict argham to parse as ranges.

    default_start : int, optional
        Default start for ranges when not specified.

    default_end : int, optional
        Default end for ranges when not specified.

    Attributes
    ----------
    number_list : list of int
        Parsed individual number values.

    range_list : list of range
        Parsed numeric ranges.

    Methods
    -------
    __contains__(value)
        Implements 'in' membership check.

    check_contained(values)
        Check membership for single or multiple values.

    Examples
    --------
    >>> arg = Argham([1, 2, 3, 4, {'start':6, 'end':10}")
    >>> 2 in arg
    True

    >>> arg.check_contained([4, 5, 6])
    [True, False, True]
    """

    def __init__(self, argham: list | dict | int | None = None, **kwargs):
        self.range_set = set()
        self.defaults = ArghamDefaults(**kwargs)
        self.min: int | None = None
        self.max: int | None = None
        self._parse_argham(argham)

    def check_contained(self, values: int | Iterable[int]) -> bool | list[bool]:
        """Check membership of values in container.

        For a single value, checks whether the value is a member
        of the container using `in` and returns a bool result.

        For multiple values given as an iterable, checks each value
        individually and returns a list of boolean results.

        Parameters
        ----------
        values: int or iterable of int
            Single value or iterable of values to check

        Returns
        -------
        bool or list of bool
            If single value, bool indicating membership.
            If multiple values, list of bool indicating membership
            for each value.

        Examples
        --------
        >>> container.check_membership(2)
        True

        >>> container.check_membership([1, 2, 3])
        [False, True, False]
        """
        if isinstance(values, int):
            return values in self
        result = []
        for value in values:
            result.append(value in self)
        return result

    def get_numbers(self) -> set[int]:
        numbers = set()
        for rng in self.range_set:
            numbers = numbers.union(set(rng))
        return numbers

    def __repr__(self) -> str:
        integers = []
        ranges = []

        for rng in self.range_set:
            if rng.stop - rng.start == 1:
                integers.append(str(rng.start))
            else:
                ranges.append((rng.start, rng.stop))
        representation_list = []
        if len(integers) > 0:
            representation_list.append(f"[{', '.join(integers)}]")
        for rng in ranges:
            representation_list.append(f"({rng[0]} - {rng[1]})")
        return ", ".join(representation_list)

    def __contains__(self, value: int):
        if self.min is None:
            return False
        if (value < self.min) or (value > self.max):  # type: ignore
            return False
        for number_range in self.range_set:
            if value in number_range:
                return True
        return False

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, int):
            if (len(self.range_set) == 1) and (__value in self.range_set.pop()):
                return True
        if isinstance(__value, range):
            if (len(self.range_set) == 1) and (self.range_set.pop() == __value):
                return True
        if isinstance(__value, Argham):
            if self.range_set == __value.range_set:
                return True
            if self.get_numbers() == __value.get_numbers():
                return True
        return False

    def __add__(self, __other: "Argham") -> "Argham":
        if self.defaults != __other.defaults:
            print(f"Warning! different defaults! {self.defaults}, {__other.defaults}")
        result = Argham()
        result.defaults = self.defaults
        result.range_set = self.range_set.union(__other.range_set)

        if (self.min is None) and (__other.min is None):
            result.min = None
        elif self.min is None:
            result.min = __other.min
        elif __other.min is None:
            result.min = self.min
        else:
            result.min = min(self.min, __other.min)

        if (self.max is None) and (__other.max is None):
            result.max = None
        elif self.max is None:
            result.max = __other.max
        elif __other.max is None:
            result.max = self.max
        else:
            result.max = max(self.max, __other.max)
        return result

    def _parse_argham(self, argham) -> None:
        if isinstance(argham, list):
            for ragham in argham:
                self._parse_argham(ragham)
        elif isinstance(argham, dict):
            self._parse_dict(argham)
        elif isinstance(argham, int):
            drng = self.defaults.default_range
            if (drng is not None) and ((argham < drng[0]) or (argham > drng[1])):
                return
            self.range_set.add(range(argham, argham + 1))
            self._update_min(argham)
            self._update_max(argham)
        else:
            pass

    def _parse_dict(self, dictionary: dict) -> None:
        if len(self.defaults.keywords) > 0:
            for word in self.defaults.keywords:
                if word in dictionary:
                    self._parse_argham(dictionary[word])
                    return
        if ("start" in dictionary) or ("end" in dictionary):
            self._parse_start_end_dict(dictionary)
        else:
            for value in dictionary.values():
                self._parse_argham(value)

    def _parse_start_end_dict(self, dictionary: dict) -> None:
        start = self.defaults.default_start
        end = self.defaults.default_end
        step = self.defaults.default_step

        if "start" in dictionary:
            start = dictionary["start"]
        if "end" in dictionary:
            end = dictionary["end"]
        if "step" in dictionary:
            step = dictionary["step"]

        if start is None:
            raise ValueError("Start must be specified")
        if end is None:
            raise ValueError("End must be specified")

        self.range_set.add(range(start, end, step))
        self._update_min(start)
        self._update_max(end - 1)

    def _update_min(self, number) -> None:
        if self.min is None:
            self.min = number
        elif number < self.min:
            self.min = number

    def _update_max(self, number) -> None:
        if self.max is None:
            self.max = number
        elif self.max < number:
            self.max = number



from pathlib import Path
import tomllib

import requests
import boto3

from .metadata_reader import Defaults, Metadata


def _get_bucket(bucket_name: str):
    with open("tokens.toml", "rb") as file:
        token = tomllib.load(file)["arvan"]
    s3_resource = boto3.resource(
        "s3",
        endpoint_url="https://s3.ir-tbz-sh1.arvanstorage.ir",
        aws_access_key_id=token["access_key"],
        aws_secret_access_key=token["secret_key"],
    )
    bucket = s3_resource.Bucket(bucket_name)  # type: ignore
    return bucket


class Maintainer:
    def __init__(self, lib_defaults: Defaults, lib_metadata: Metadata) -> None:
        self.bucket_name = lib_defaults.bucket_address.split("/")[-1]
        self.bucket = _get_bucket(self.bucket_name)
        self.lib_defaults = lib_defaults
        self.lib_metadata = lib_metadata

    def upload_raw_files(self) -> None:
        for year in self.lib_metadata.raw_files.keys():
            files = self.lib_metadata.raw_files[year]
            for file in files.get("compressed_files", []):
                file_str_path = f"{year}/{file['name']}"
                url = f"{self.lib_defaults.online_dirs.compressed}/{file_str_path}"
                file_path = self.lib_defaults.dirs.compressed.joinpath(file_str_path)
                self._upload_file_to_online_directory(file_path, url)
            for file in files.get("unpacked_files", []):
                file_str_path = f"{year}/{file['name']}"
                url = f"{self.lib_defaults.online_dirs.unpacked}/{file_str_path}"
                file_path = self.lib_defaults.dirs.unpacked.joinpath(file_str_path)
                self._upload_file_to_online_directory(file_path, url)

    def upload_cleaned_files(self) -> None:
        for file_path in self.lib_defaults.dirs.cleaned.iterdir():
            url = f"{self.lib_defaults.online_dirs.cleaned}/{file_path.name}"
            self._upload_file_to_online_directory(file_path, url)

    def ckeck_if_up_to_date(self, file_path: Path, url: str) -> bool:
        response = requests.head(url, timeout=10)
        try:
            online_file_size = int(response.headers["Content-Length"])
        except KeyError:
            online_file_size = 0
        local_file_size = file_path.stat().st_size
        return online_file_size == local_file_size

    def _upload_file_to_online_directory(self, file_path: Path, url: str) -> None:
        if self.ckeck_if_up_to_date(file_path, url):
            return
        url_parts = url.split("/")
        bucket_index = url_parts.index(self.bucket_name)
        key = "/".join(url_parts[bucket_index + 1 :])
        with open(file_path, "rb") as file:
            self.bucket.put_object(ACL="public-read", Body=file, Key=key)
